library geiger_localstorage;

import 'package:geiger_localstorage/geiger_localstorage.dart';
import 'package:geiger_localstorage/src/db/data/switchable_boolean.dart';
import 'package:geiger_localstorage/src/serialization/byte_sink.dart';
import 'package:geiger_localstorage/src/serialization/byte_stream.dart';
import 'package:geiger_localstorage/src/serialization/serializer_helper.dart';

/// <p>The implementation of the node interface.</p>
///
/// <p>This Class denotes one node containing sub-nodes in a tree-like structure. Each node may have
/// n children. Each node may be a skeleton-only node (contains only name and a reference to a
/// mapper), or may be materialized (contains all data). Typically when fetching a node, the node
/// is materialized but its sub-nodes are skeleton-only nodes. All skeleton nodes materialize
/// automatically if their data is accessed.</p>
class NodeImpl implements Node {
  static const int serialversionUID = 11239348938;

  /// an indicator whether the current object is a skeleton
  final SwitchableBoolean _skeleton = SwitchableBoolean(false);

  /// Contains the mapper for a skeleton to fetch any subsequent  data
  StorageController? _controller;

  /// contains the ordinals of a node
  final Map<Field, String> _ordinals = <Field, String>{Field.owner: ''};

  /// contains the key/value pairs of a node
  final Map<String, NodeValue> _values = {};

  /// Holds all child nodes as tuples, where the name is used as a key and
  /// the value is of type StorageNode
  final Map<String, Node> _childNodes = {};

  /// <p>Constructor creating a skeleton node.</p>
  ///
  /// @param path       the path of the node
  /// @param controller the controller to fetch the full node
  NodeImpl.createSkeleton(String path, this._controller) {
    _skeleton.set(true);
    try {
      set(Field.path, path);
    } on StorageException {
      throw Exception('Oops.... this should not happen... contact developer');
    }
  }

  NodeImpl.fromNode(Node node) {
    update(node);
  }

  /// Create a empty node for the given path.
  ///
  /// @param path the node path
  /// @param isTombstone true if the node is a tombstone node
  /// @param visibility the visibility of the node or null if default
  /// @param nodeValues the node values to be stored or null if none
  /// @param childNodes the child nodes to be included or null if none
  factory NodeImpl.fromPath(String path,
      {bool? isTombstone,
      Visibility? visibility,
      List<NodeValue>? nodeValues,
      List<Node>? childNodes}) {
    var node = NodeImpl(getNameFromPath(path)!, getParentFromPath(path));
    try {
      if (isTombstone != null) {
        node.set(Field.tombstone, isTombstone ? 'true' : 'false');
      }
      if (visibility != null) {
        node.set(Field.visibility, visibility.toValueString());
      }
      if (nodeValues != null) {
        for (var nv in nodeValues) {
          node.addValue(nv);
        }
      }
      if (childNodes != null) {
        for (var n in childNodes) {
          node.addChildNode(n);
        }
      }
    } on StorageException {
      throw Exception('Oops.... this should not happen... contact developer');
    }
    return node;
  }

  /// <p>creates a fully fledged new empty node.</p>
  ///
  /// @param name   the name for the node
  /// @param parent the parent of the node (may be null if root node is the parent)
  /// @param vis    visibility of the node
  NodeImpl(String name, [String? parent, Visibility vis = Visibility.red]) {
    if (parent == null) {
      parent = getParentFromPath(name);
      name = getNameFromPath(name) ?? '';
    }
    try {
      set(Field.path, '$parent${GenericController.pathDelimiter}$name');
      set(Field.visibility, vis.toString());
    } on StorageException {
      throw Exception('Oops.... this should not happen... contact developer');
    }
    _skeleton.set(false);
  }

  /// <p>Converts current node into a materialized node from a skeleton.</p>
  void init() {
    if (skeleton) {
      update(controller!.get(path!));
      _skeleton.set(false);
      _controller = null;
    }
  }

  /// <p>Returns the name part from a fully qualified node path.</p>
  ///
  /// @param path the fully qualified path of which the name is extracted
  /// @return the name part of the path
  static String? getNameFromPath(String? path) {
    if (path == null) {
      return null;
    }
    return path
        .substring(path.lastIndexOf(GenericController.pathDelimiter) + 1);
  }

  /// <p>Returns the fully qualified node path of the parental node.</p>
  ///
  /// @param path the fully qualified path of which the parental node is extracted
  /// @return the fully qualified path to the parental node
  static String? getParentFromPath(String? path) {
    if (path == null) {
      return null;
    }
    if (!path.contains(GenericController.pathDelimiter)) {
      return '';
    }
    return path.substring(0, path.lastIndexOf(GenericController.pathDelimiter));
  }

  @override
  NodeValue? getValue(String key) {
    init();
    NodeValue? ret = _values[key];
    if (ret != null) {
      ret = ret.deepClone();
    }
    return ret;
  }

  @override
  NodeValue updateValue(NodeValue value) {
    init();
    NodeValue? ret = getValue(value.key);
    if (ret == null) {
      throw StorageException(
          'Value ' + value.key + ' not found in node ' + (name ?? 'null'));
    }
    _values[value.key] = value;

    return ret;
  }

  @override
  void addValue(NodeValue value) {
    init();
    if (getValue(value.key) != null) {
      throw StorageException(
          'value "${value.key}" does already exist in the node "$path"');
    }
    _values[value.key] = value;
  }

  @override
  void addOrUpdateValue(NodeValue value) {
    _values[value.key] = value;
  }

  @override
  NodeValue? removeValue(String key) {
    init();
    return _values.remove(key);
  }

  @override
  void addChild(Node node) {
    init();
    String key = node.name!;
    if (!_childNodes.containsKey(key)) {
      _childNodes[key] = node;
    }
  }

  @override
  String get owner {
    try {
      return get(Field.owner) ?? '';
    } on StorageException {
      throw Exception('Oops.... this should not happen... contact developer');
    }
  }

  @override
  set owner(String newOwner) {
    try {
      set(Field.owner, newOwner);
    } on StorageException {
      throw Exception('Oops.... this should not happen... contact developer');
    }
  }

  @override
  String? get name {
    try {
      return getNameFromPath(get(Field.path));
    } on StorageException {
      throw Exception('Oops.... this should not happen... contact developer');
    }
  }

  @override
  String? get parentPath {
    try {
      return getParentFromPath(get(Field.path));
    } on StorageException {
      throw Exception('Oops.... this should not happen... contact developer');
    }
  }

  @override
  String? get path {
    try {
      return get(Field.path);
    } on StorageException {
      throw Exception('Oops.... this should not happen... contact developer');
    }
  }

  @override
  Visibility get visibility {
    try {
      var rawVisibility = get(Field.visibility);
      return VisibilityExtension.valueOf(rawVisibility!) ?? Visibility.red;
    } on StorageException {
      throw Exception('Oops.... this should not happen... contact developer');
    }
  }

  @override
  set visibility(Visibility newVisibility) {
    try {
      set(Field.visibility, newVisibility.toString());
    } on StorageException {
      throw Exception('Oops.... this should not happen... contact developer');
    }
  }

  @override
  Map<String, NodeValue> getValues() {
    return Map.from(_values);
  }

  /// <p>Gets an ordinal field of the node.</p>
  /// @param field the ordinal field to get
  /// @return the currently set value
  /// @throws StorageException if a field does not exist
  String? get(Field field) {
    if (((field != Field.path) && (field != Field.name)) &&
        (field != Field.tombstone)) {
      init();
    }

    switch (field) {
      case Field.owner:
      case Field.path:
      case Field.visibility:
      case Field.lastModified:
      case Field.expiry:
      case Field.tombstone:
        return _ordinals[field];
      case Field.name:
        return getNameFromPath(_ordinals[Field.path]);
      default:
        throw StorageException('unable to fetch field ' + field.toString());
    }
  }

  /// <p>Sets an ordinal field of the node.</p>
  ///
  /// @param field the ordinal field to be set
  /// @param value the new value
  /// @return the previously set value
  /// @throws StorageException if a field does not exist
  String? set(Field field, String? value) {
    if (field != Field.path) {
      init();
    }

    var current = _ordinals[field];

    if (field != Field.lastModified &&
        ((current != null && current != value) ||
            (current == null && value != null))) {
      touch();
    }
    switch (field) {
      case Field.owner:
      case Field.path:
      case Field.visibility:
      case Field.lastModified:
      case Field.expiry:
        if (value == null) {
          return _ordinals.remove(field);
        }
        var prev = _ordinals[field];
        _ordinals[field] = value;
        return prev;
      case Field.tombstone:
        var prev = _ordinals[field] ?? 'false';
        _ordinals[field] = ('true' == value) ? 'true' : 'false';
        return prev;
      default:
        throw StorageException('unable to set field ' + field.toString());
    }
  }

  /// <p>Checks if the current node is marked as TOMBSTONE (old deleted node).</p>
  /// @return true if node is a tombstone
  @override
  bool get tombstone {
    try {
      return 'true' == get(Field.tombstone);
    } on StorageException {
      throw Exception('OOPS! Unexpected exception... please contact developer');
    }
  }

  void addChildNode(Node n) {
    _childNodes[n.name!] = n;
  }

  @override
  void removeChild(String name) {
    _childNodes.remove(name);
  }

  @override
  Map<String, Node> getChildren() {
    init();
    Map<String, Node> ret = {};
    for (var entry in _childNodes.entries) {
      ret[entry.key] = entry.value.deepClone();
    }
    return ret;
  }

  @override
  Node? getChild(String name) {
    init();
    return _childNodes[name];
  }

  @override
  String getChildNodesCsv() {
    init();
    if (_childNodes.isEmpty) {
      return '';
    }
    var csv = '';
    for (var s in _childNodes.keys) {
      csv += s + ',';
    }
    return csv.substring(0, csv.length - 1);
  }

  @override
  bool get skeleton {
    return _skeleton.get();
  }

  @override
  StorageController? get controller {
    return _controller;
  }

  @override
  set controller(StorageController? controller) {
    _controller = controller;
  }

  @override
  bool operator ==(Object other) => equals(other);

  @override
  bool equals(Object? object) {
    if (object == null || object is! NodeImpl) {
      return false;
    }
    var n2 = object;
    if (!skeleton || (skeleton && (!n2.skeleton))) {
      try {
        init();
        n2.init();
      } on StorageException {
        // I do not care if init fails....
      }
      if (_ordinals.length != n2._ordinals.length) {
        return false;
      }
      for (var e in n2._ordinals.entries) {
        try {
          if (!(e.value == get(e.key))) {
            return false;
          }
        } on Exception {
          throw Exception(
              'Oops.... this should not happen... contact developer');
        }
      }
      if (_values.length != n2._values.length) {
        return false;
      }
      for (var e in _values.entries) {
        try {
          if (!(e.value.equals(n2.getValue(e.key)))) {
            return false;
          }
        } on StorageException {
          return false;
        }
      }
      if (_childNodes.length != n2._childNodes.length) {
        return false;
      }
      for (var n in _childNodes.keys) {
        if (n2._childNodes[n] == null) {
          return false;
        }
      }
    } else {
      if (!(path == n2.path)) {
        return false;
      }
      if (controller != n2.controller) {
        return false;
      }
    }
    return true;
  }

  @override
  void update(Node n2, {bool deepClone = true}) {
    _controller = n2.controller;
    _skeleton.set(n2.skeleton);
    var path = (n2 as NodeImpl)._ordinals[Field.path];
    if (path != null) {
      _ordinals[Field.path] = path;
    }
    if (!n2.skeleton) {
      // synchronized(ordinals, {
      _ordinals.clear();
      for (var e in n2._ordinals.entries) {
        _ordinals[e.key] = e.value;
      }
      // });
      // synchronized(values, {
      _values.clear();
      for (var e in n2._values.entries) {
        _values[e.key] = e.value.deepClone();
      }
      //});

// synchronized(childNodes, {
      _childNodes.clear();

      for (var e in n2.getChildren().entries) {
        if (deepClone) {
          _childNodes[e.key] = e.value.deepClone();
        } else {
          _childNodes[e.key] =
              NodeImpl.createSkeleton(e.value.path!, controller);
        }
      }
      // });
    }

    var lastModified = n2._ordinals[Field.lastModified];
    if (lastModified != null) {
      _ordinals[Field.lastModified] = lastModified;
    }
  }

  @override
  Node deepClone() {
    return NodeImpl.fromNode(this);
  }

  @override
  Node shallowClone() {
    var ret = NodeImpl.fromPath(path!);
    ret.update(this, deepClone: false);
    return ret;
  }

  void touch() {}

  @override
  String toString() {
    var sb = StringBuffer();
    sb.write(path);
    sb.write('[');
    if (skeleton) {
      sb.write('{<skeletonized>}');
      sb.write(']{');
      sb.write('\n');
    } else {
      sb.write('owner=');
      sb.write(owner);
      sb.write(';vis=');
      sb.write(visibility);
      sb.write(']{');
      sb.write('\n');
    }
    var i = 0;
    if (skeleton) {
      sb.write('{<skeletonized>}');
    } else {
      for (var e in _values.entries) {
        if (i > 0) {
          sb.write(', ');
          sb.write('\n');
        }
        sb.write(e.value.toString('  '));
        i++;
      }
      sb.write('\n');
      sb.write('}');
    }
    return sb.toString();
  }

  @override
  int get hashCode => toString().hashCode;

  @override
  void toByteArrayStream(ByteSink out) {
    SerializerHelper.writeLong(out, serialversionUID);
    SerializerHelper.writeInt(out, skeleton ? 1 : 0);
    SerializerHelper.writeString(out, path);
    SerializerHelper.writeInt(out, _ordinals.length - 1);
    for (MapEntry<Field, String> e in _ordinals.entries) {
      if (e.key != Field.path) {
        SerializerHelper.writeString(out, e.key.toString());
        SerializerHelper.writeString(out, e.value);
      }
    }

    if (!skeleton) {
      SerializerHelper.writeInt(out, _values.length);
      for (MapEntry<String, NodeValue> e in _values.entries) {
        SerializerHelper.writeString(out, e.key);
        e.value.toByteArrayStream(out);
      }
    }

    SerializerHelper.writeInt(out, _childNodes.length);
    for (MapEntry<String, Node> e in _childNodes.entries) {
      SerializerHelper.writeString(out, e.key);
      e.value.toByteArrayStream(out);
    }

    SerializerHelper.writeLong(out, serialversionUID);
  }

  /// <p>Deserializes a NodeValue from a byteStream.</p>
  /// @param in the stream to be read
  /// @return the deserialized NodeValue
  /// @throws IOException if an exception happens deserializing the stream
  static Future<NodeImpl> fromByteArrayStream(ByteStream in_) async {
    if (await SerializerHelper.readLong(in_) != serialversionUID) {
      throw Exception('failed to parse NodeImpl (bad stream?)');
    }
    var skel = (await SerializerHelper.readInt(in_) == 1);
    var n = NodeImpl(await SerializerHelper.readString(in_) ?? '');
    if (skel) {
      n.controller = GenericController.getDefault();
    }
    int counter = await SerializerHelper.readInt(in_);
    for (var i = 0; i < counter; i++) {
      n._ordinals[
              Field.fromString(await SerializerHelper.readString(in_) ?? '')??Field.owner] =
          await SerializerHelper.readString(in_) ?? '';
    }
    if (!skel) {
      counter = await SerializerHelper.readInt(in_);
      for (var i = 0; i < counter; i++) {
        n._values[await SerializerHelper.readString(in_) ?? ''] =
            await NodeValueImpl.fromByteArrayStream(in_);
      }
      counter = await SerializerHelper.readInt(in_);
      for (var i = 0; i < counter; i++) {
        n._childNodes[await SerializerHelper.readString(in_) ?? ''] =
            await NodeImpl.fromByteArrayStream(in_);
      }
    }
    if (await SerializerHelper.readLong(in_) != serialversionUID) {
      throw Exception('failed to parse NodeImpl (bad stream end?)');
    }
    return n;
  }
}
