import 'package:geiger_localstorage/geiger_localstorage.dart';

/// <p>A non-persisting dummy mapper for test purposes.</p>
class DummyMapper extends AbstractMapper {
  final Map<String, Node> nodes = {};
  StorageController? controller;

  @override
  void setController(StorageController controller) {
    this.controller = controller;
  }

  @override
  Node get(String path) {
    checkPath(path);
    getSanity(path);
    var ret = nodes[path];
    if (ret == null) {
      throw StorageException('Node not found');
    }
    if (ret.tombstone) {
      Node returnNode = NodeImpl.fromPath(path, isTombstone: true);
      returnNode.visibility = ret.visibility;
      return returnNode;
    }
    return ret.deepClone();
  }

  @override
  void add(Node node) {
    checkPath(node.path!);
    // synchronized(nodes, {
    if (nodes[node.path] != null) {
      throw StorageException('Node does already exist');
    }
    if (node.skeleton) {
      throw StorageException('Skeleton nodes cannot be added.');
    }
    if ((node.parentPath != null) && (!('' == node.parentPath))) {
      if (nodes[node.parentPath] == null) {
        throw StorageException(
            ('Parent node "' + (node.parentPath!)) + '" does not exist');
      }
      nodes[node.parentPath]?.addChild(node);
    }
    nodes[node.path!] = node.shallowClone();
    // });
  }

  @override
  void update(Node node) {
    checkPath(node.path!);
    // synchronized(nodes, {
    if ((!('' == node.parentPath)) && (nodes[node.parentPath] == null)) {
      throw StorageException('Parent node ${node.parentPath} does not exist');
    }
    nodes[node.path]?.update(node);
    // });
  }

  @override
  void rename(String oldPath, String newPathOrName) {
    checkPath(oldPath);
    checkPath(newPathOrName);
    // synchronized(nodes, {
    var oldNode = nodes[oldPath];
    if (oldNode == null) {
      throw StorageException(
          'Old node "$oldPath" does not exist.. teherefore unable to rename to "$newPathOrName"');
    }
    var newNode = NodeImpl.fromPath(newPathOrName);
    newNode.owner = oldNode.owner;
    newNode.visibility = oldNode.visibility;
    for (var nv in oldNode.getValues().values) {
      newNode.addValue(nv);
    }
    add(newNode);
    for (var n in oldNode.getChildren().values) {
      rename(
          n.path!, (newNode.path! + GenericController.pathDelimiter) + n.name!);
    }
    delete(oldNode.path!);
    // });
  }

  @override
  Node delete(String path) {
    // synchronized(nodes, {
    var node = nodes[path];
    if (node == null) {
      throw StorageException('unable to delete node "$path"... does not exist');
    }
    if (!('' == node.getChildNodesCsv())) {
      throw StorageException(
          'Node does have children... cannot remove ' + path);
    }
    nodes.remove(node);
    Node tombstone = NodeImpl.fromPath(node.path!, isTombstone: true);
    tombstone.visibility = node.visibility;
    nodes[node.path!] = tombstone;
    if ((node.parentPath != null) && (!('' == node.parentPath))) {
      nodes[node.parentPath]!.removeChild(node.name!);
    }
    return node;
    // });
  }

  @override
  NodeValue getValue(String path, String key) {
    var ret = get(path);
    if (ret.tombstone) {
      throw StorageException('node does not exist');
    }
    return ret.getValues()[key]!;
  }

  @override
  List<Node> search(SearchCriteria criteria) {
    var l = List<Node>.empty(growable: true);
    for (var e in nodes.entries) {
      if (criteria.evaluate(e.value)) {
        l.add(e.value);
      }
    }
    return l;
  }

  @override
  void close() {}

  @override
  void flush() {}

  @override
  void zap() {
    // synchronized(nodes, {
    nodes.clear();
    // });
  }
}
